/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.java2d;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

/** This class captures the information needed to decribe the textual
 * annotation (a finite number of text strings at given positions) of a scene
 * graph node. */

public class Annotation extends Point2D.Double {
	
	private transient int hashCode;
	
	/** The anchor point is at the lower left corner of the bounding box of the
	 * text. */
	
	public static final int SOUTHWEST = 0;
	
	/** The anchor point is in the center of the bottom edge of the bounding
	 * box of the text. */
	
	public static final int SOUTH = 1;
	
	/** The anchor point is at the lower right corner of the bounding box of
	 * the text. */
	
	public static final int SOUTHEAST = 2;
	
	/** The anchor point is in the center of the right edge of the bounding
	 * box of the text. */
	
	public static final int EAST = 3;
	
	/** The anchor point is at the upper right corner of the bounding box of the
	 * text. */
	
	public static final int NORTHEAST = 4;
	
	/** The anchor point is in the center of the top edge of the bounding
	 * box of the text. */
	
	public static final int NORTH = 5;
	
	/** The anchor point is at the upper left corner of the bounding box of the
	 * text. */
	
	public static final int NORTHWEST = 6;
	
	/** The anchor point is in the center of the left edge of the bounding
	 * box of the text. */
	
	public static final int WEST = 7;
	
	/** The anchor point is at the center of the bounding box of the
	 * text. */
	
	public static final int CENTER = 8;
	
	String text;
	int anchor = SOUTHWEST;
	
	/** Creates an annotation with no text strings. */
	
	public Annotation() {}
	
	public Annotation(String text, double x, double y, int anchor) {
		setLocation(x,y);
		this.text = text;
		this.anchor = anchor;
	}
	
	public Annotation(Annotation a) {
		this(a.getText(), a.x, a.y, a.getAnchor());
	}
		
	/**
	 * Returns the text.
	 */
	public String getText() {
		return text;
	}
	
	/**
	 * Sets the text.
	 */
	public void setText(String s) {
		text = s;
	}
	
	public void setX(double x) {
		this.x = x;
	}

	public void setY(double y) {
		this.y = y;
	}
	
	/** Tells where anchor point the text is located on the
	 * bounding box of the text , possibly with an additional offset in the
	 * indicated direction. The possible values are given by the static fields
	 * CENTER, NORTH, NORTHWEST ... of this class. */
	
	public int getAnchor() {
		return anchor;
	}
	
	/** Sets the direction where the text should be drawn
	 * relative to the point given by its coordinates. The possible values are
	 * indicated by the static fields Annotation.CENTER, Annotation.NORTH
	 * etc. */
	
	public void setAnchor(int a) {
		anchor = a;
	}
	
	public int hashCode() {
		return hashCode;
	}
	
	public boolean equals(Object o) {
		return this == o;
	}

	static void writeOffsets(
			int margin,
			Rectangle2D.Double r,
			int anchor,
			double[] offsets
	) {			
		switch (anchor) {
		case Annotation.SOUTHWEST:
			offsets[0] = margin;
			offsets[1] = -(r.height + r.y) - margin;
			break;
		case Annotation.WEST:
			offsets[0] = margin;
			offsets[1] = -(.5 * r.height + r.y);
			break;
		case Annotation.NORTHWEST:
			offsets[0] = margin;
			offsets[1] = -r.y + margin;
			break;
		case Annotation.NORTH:
			offsets[0] = -.5 * r.width;
			offsets[1] = -r.y + margin; 
			break;
		case Annotation.NORTHEAST:
			offsets[0] = -r.width - margin;
			offsets[1] = -r.y + margin;
			break;
		case Annotation.EAST:
			offsets[0] = -r.width - margin;
			offsets[1] = -(.5 *r.height + r.y);
			break;
		case Annotation.SOUTHEAST:
			offsets[0] = -r.width-margin;
			offsets[1] = -(r.height + r.y) - margin;
			break;
		case Annotation.SOUTH:
			offsets[0] = -.5 * r.width;
			offsets[1] = -(r.height + r.y) - margin;
			break;	
		case Annotation.CENTER:
			offsets[0] = -.5 * r.width;
			offsets[1] = -(.5 *r.height + r.y);
			break;
		}
	}
}