/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.java2d;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.font.FontRenderContext;
import java.awt.geom.AffineTransform;
import java.awt.geom.Ellipse2D;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.Collection;
import java.util.List;

import javax.swing.Icon;


class SceneAttributes implements Cloneable {
	Paint paint = Color.yellow;
	Paint outlinePaint = Color.black;
	Paint pointPaint = Color.red;
	Paint annotationPaint = Color.blue;
	Paint highlightPaint = new Color(255,255,0,80);
	Paint highlightOutlinePaint = new Color(80,80,80,255);
	Paint highlightPointPaint = new Color(128,128,0,80);
	Paint highlightAnnotationPaint = new Color(0,0,128,80);
	Stroke stroke = new BasicStroke();
	Stroke pointStroke = new BasicStroke();
	Shape pointShape = new Ellipse2D.Double(-3,-3,6,6);
	Integer annotationMargin = new Integer(5);
	Font font = new Font("Lucida", 0, 10);
	Boolean filled = Boolean.TRUE;
	Boolean outlined = Boolean.TRUE;
	Boolean pointFilled = Boolean.TRUE;
	Boolean pointOutlined = Boolean.TRUE;
	Boolean highlighted = Boolean.FALSE;
	AffineTransform tr;
	AffineTransform inverseTr;
	
	SceneAttributes(AffineTransform t, AffineTransform i) {
		tr = t;
		inverseTr = i;
	}
	
	void setupForPick(SceneComponent c) {
		Stroke componentStroke = c.getStroke();
		Shape componentPointShape = c.getPointShape();
		Integer componentAnnotationMargin = c.getAnnotationMargin();
		Font componentFont = c.getFont();
		AffineTransform componentTransform = c.getTransform();
		if (componentAnnotationMargin != null) {
			annotationMargin = componentAnnotationMargin;
		}
		if (componentFont != null) {
			font = componentFont;
		}
		if (componentStroke != null) {
			stroke = componentStroke;
		}
		if (componentPointShape != null) {
			pointShape = componentPointShape;
		}
		if (componentTransform != null) {
			tr = (AffineTransform)tr.clone();
			tr.concatenate(componentTransform);
			try {
				inverseTr = tr.createInverse();
			} catch (NoninvertibleTransformException f) {
				System.err.println("transform not invertible");
			}
		}
	}
	
	void setupForPaint(SceneComponent c) {
		Paint componentPaint = c.getPaint();
		Paint componentOutlinePaint = c.getOutlinePaint();
		Paint componentPointPaint = c.getPointPaint();
		Paint componentAnnotationPaint = c.getAnnotationPaint();
		Paint componentHighlightPaint = c.getHighlightPaint();
		Paint componentHighlightOutlinePaint = c.getHighlightOutlinePaint();
		Paint componentHighlightPointPaint = c.getHighlightPointPaint();
		Paint componentHighlightAnnotationPaint = c.getHighlightAnnotationPaint();
		Stroke componentStroke = c.getStroke();
		Stroke componentPointStroke = c.getPointStroke();
		Shape componentPointShape = c.getPointShape();
		Integer componentAnnotationMargin = c.getAnnotationMargin();
		Font componentFont = c.getFont();
		AffineTransform componentTransform = c.getTransform();
		Boolean componentFilled = c.getFilled();
		Boolean componentOutlined = c.getOutlined();
		Boolean componentPointFilled = c.getPointFilled();
		Boolean componentPointOutlined = c.getPointOutlined();
		Boolean componentHighlighted = c.getHighlighted();

		if (componentPaint != null) {
			paint = componentPaint;
		}
		if (componentOutlinePaint != null) {
			outlinePaint = componentOutlinePaint;
		}
		if (componentPointPaint != null) {
			pointPaint = componentPointPaint;
		}
		if (componentAnnotationPaint != null) {
			annotationPaint = componentAnnotationPaint;
		}
		if (componentHighlightPaint != null) {
			highlightPaint = componentHighlightPaint;
		}
		if (componentHighlightOutlinePaint != null) {
			highlightOutlinePaint = componentHighlightOutlinePaint;
		}
		if (componentHighlightPointPaint != null) {
			highlightPointPaint = componentHighlightPointPaint;
		}
		if (componentHighlightAnnotationPaint != null) {
			highlightAnnotationPaint = componentHighlightAnnotationPaint;
		}
		if (componentStroke != null) {
			stroke = componentStroke;
		}
		if (componentPointStroke != null) {
			pointStroke = componentPointStroke;
		}
		if (componentPointShape != null) {
			pointShape = componentPointShape;
		}
		if (componentAnnotationMargin != null) {
			annotationMargin = componentAnnotationMargin;
		}
		if (componentFont != null) {
			font = componentFont;
		}
		if (componentFilled != null) {
			filled = componentFilled;
		}
		if (componentOutlined != null) {
			outlined = componentOutlined;
		}
		if (componentPointFilled != null) {
			pointFilled = componentPointFilled;
		}
		if (componentPointOutlined != null) {
			pointOutlined = componentPointOutlined;
		}
		if (componentHighlighted != null) {
			highlighted = componentHighlighted;
		}
		if (componentTransform != null) {
			tr = (AffineTransform)tr.clone();
			tr.concatenate(componentTransform);
			try {
				inverseTr = tr.createInverse();
			} catch (NoninvertibleTransformException f) {
				System.err.println("transform not invertible");
			}
		}
	}

	Rectangle2D.Double getBounds2D(SceneComponent c) {
		Rectangle2D r = null;
		Shape shape = c.getShape();
		Point2D.Double p = new Point2D.Double();
		final Collection<Point2D.Double> points = c.getPoints();
		if (shape != null) {
			Shape transformedShape = tr.createTransformedShape(shape);
			r = transformedShape.getBounds2D();
		}
		Rectangle2D.Double s = null;
		if (!points.isEmpty()) {
			double xMin = Double.MAX_VALUE;
			double xMax = -Double.MAX_VALUE;
			double yMin = Double.MAX_VALUE;
			double yMax = -Double.MAX_VALUE;
			for (Point2D.Double q : points) {
				p.setLocation(q);
				tr.transform(p, p);
				if (p.x < xMin) xMin = p.x;
				if (p.x > xMax) xMax = p.x;
				if (p.y < yMin) yMin = p.y;
				if (p.y > yMax) yMax = p.y;
			}
			s = new Rectangle2D.Double(xMin,yMin,xMax-xMin,yMax-yMin);
		}
		if (s!= null) {
			if (r == null) {
				r = s;
			} else {
				Rectangle2D.union(r, s, r);
			}
		}
		if (r != null) {
			Rectangle2D.Double result = new Rectangle2D.Double();
			result.setRect(r);
			return result;
		} else {
			return null;
		}
	}
	
	void draw(
			Component viewer,
		Graphics2D g,
		SceneComponent c
	) {
		AffineTransform t=g.getTransform();
		g.transform(tr);

		Shape shape = c.getShape();
		final Collection<Point2D.Double> points = c.getPoints();
		if (shape != null) {
			if (filled.booleanValue()) {
				g.setPaint(highlighted.booleanValue() ? highlightPaint : paint);
				g.fill(shape);
			}
			
			if (outlined.booleanValue()) {
				Shape transformedShape = tr.createTransformedShape(shape);
				Shape outlineScreen =
					stroke.createStrokedShape(transformedShape);
				Shape outlineUser =
					inverseTr.createTransformedShape(outlineScreen);
				g.setPaint(highlighted.booleanValue() ? highlightOutlinePaint : outlinePaint);
				g.fill(outlineUser);
			}
		}
		if (!points.isEmpty()) {
			final Point2D.Double p = new Point2D.Double();
			for (Point2D.Double q : points) {
				p.setLocation(q);
				
				tr.transform(p,p);
				AffineTransform a = AffineTransform.getTranslateInstance(p.x,p.y);
				Shape pointShapeTmp = a.createTransformedShape(
						pointShape
				);
				Shape pointShapeUser = inverseTr.createTransformedShape(pointShapeTmp);
				if (pointFilled.booleanValue()) {
					g.setPaint(highlighted.booleanValue() ? highlightPointPaint : pointPaint);
					g.fill(pointShapeUser);
				}
				if (pointOutlined.booleanValue()) {
					g.setPaint(highlighted.booleanValue() ? highlightOutlinePaint : outlinePaint);
					Shape outlineScreen = pointStroke.createStrokedShape(pointShapeTmp);
					g.fill(inverseTr.createTransformedShape(outlineScreen));
				}
			}
		}
		g.setTransform(t);
		List<Icon2D> icons = c.getIcons();
		Point2D.Double p = new Point2D.Double();
		
		if (!icons.isEmpty()) {
			g.setPaint(highlighted.booleanValue() ? highlightAnnotationPaint : annotationPaint);
			g.setFont(font);
			int margin = annotationMargin.intValue();
			Rectangle2D.Double rectangle = new Rectangle2D.Double();
			final double[] offsets = new double[2];
			for (Icon2D icon2D : icons) {
				p.x = icon2D.getX();
				p.y = icon2D.getY();
				tr.transform(p,p);
				Icon icon = icon2D.getIcon();
				rectangle.setRect(0,0,icon.getIconWidth(),icon.getIconHeight());
				Annotation.writeOffsets(
						margin,
						rectangle,
						icon2D.getAnchor(),
						offsets
				);
				icon.paintIcon(
						viewer,
						g,
						(int)(p.x + offsets[0]),
						(int)(p.y + offsets[1])
				);
			}
		}
		List<Annotation> annotations = c.getAnnotations();
		if (!annotations.isEmpty()) {
			g.setPaint(highlighted.booleanValue() ? highlightAnnotationPaint : annotationPaint);
			g.setFont(font);
			int margin = annotationMargin.intValue();
			FontRenderContext frc = g.getFontRenderContext();
			Rectangle2D.Double rectangle = new Rectangle2D.Double();
			for (Annotation a : annotations) {
				p.x = a.getX();
				p.y = a.getY();
				tr.transform(p,p);
				String text=a.getText();
				if (text != null) {
					rectangle.setRect(font.getStringBounds(text,frc));
					double[] offsets = new double[2];
					Annotation.writeOffsets(
							margin,
							rectangle,
							a.anchor,
							offsets
					);
					g.drawString(
							text,
							(float)(p.x + offsets[0]),
							(float)(p.y + offsets[1])
					);
				}
			}
		}
	}

	SceneAttributes copy() {
		try {
			return (SceneAttributes) clone();
		} catch (CloneNotSupportedException e) {
			throw new Error();
		}
	}
	
	static Rectangle2D.Double getBounds2D(SceneComponent c, SceneAttributes a) {
		Rectangle2D.Double r = null;
		if (c.isVisible()) {
			a = a.copy();
			a.setupForPaint(c);
			r = a.getBounds2D(c);
			for (int i = 0; i < c.getChildCount(); i++) {
				Rectangle2D.Double s = getBounds2D(c.getChild(i), a);
				if (s != null) {
					if (r == null) {
						r = s;
					} else {
						Rectangle2D.union(r, s, r);
					}
				}
			}
		}
		return r;
	}
}
