/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.java2d;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.beans.Expression;
import java.util.LinkedList;
import java.util.List;

import javax.swing.Icon;
import javax.swing.ImageIcon;

/** <p>An implementation of the <code>SceneNode</code> interface that provides
 * bean conform "get" and "set" methods for all properties. Properties that are
 * objects are not copied but stored as references, so changes made to them
 * after using them to set a property will affect the behaviour of the
 * <code>SceneComponent</code>.</p>
 * 
 * <p>Child nodes can be
 * added and removed. </p>
 * 
 * <p>A <code>SceneComponent</code> registers as an
 * <code>AppearanceChangeListener</code> with all of its children and reacts to
 * appearance changes of its children by notifiying its own
 * <code>appearanceChange</code>
 * listeners that its own appearance (which is meant to include those of its
 * children) has changed.</p>
 * 
 * <p>Another object can register as a <code>DragListener</code> in order to
 * be notified
 * when this node is dragged. Note that dragging of a child node does not
 * automatically lead to the parent receiving drag events. If one wants
 * dragging of a children to affect the parent, one has to register an
 * appropriate <code>DragListener</code> with the child. </p>
 * */

public class SceneComponent implements AppearanceChangeListener {

	private Shape shape;
	private List<Point2D.Double> points = new LinkedList<Point2D.Double>();
	private List<Annotation> annotations = new LinkedList<Annotation>();
	private List<Icon2D> icons = new LinkedList<Icon2D>();
	private Paint paint;
	private Paint outlinePaint;
	private Paint pointPaint;
	private Paint annotationPaint;
	private Stroke stroke;
	private Stroke pointStroke;
	private Shape pointShape;
	private Integer annotationMargin;
	private Font font;
	private Boolean filled;
	private Boolean outlined;
	private Boolean pointFilled;
	private Boolean pointOutlined;
	private Boolean annotated;
	private boolean areaDragEnabled;
	private boolean outlineDragEnabled;
	private boolean pointDragEnabled;
	private boolean iconDragEnabled;
	private boolean annotationDragEnabled;
	private boolean mouseOverEnabled;
	private Paint highlightPaint;
	private Paint highlightPointPaint;
	private Paint highlightOutlinePaint;
	private Paint highlightAnnotationPaint;
	private Boolean highlighted;
	private boolean visible = true;
	private String name;
	private AffineTransform transform;
	private transient List<SceneComponent> children =
		new LinkedList<SceneComponent>();
	private transient List<DragListener> dragListeners =
		new LinkedList<DragListener>();
	private transient List<MouseOverListener> mouseOverListeners =
		new LinkedList<MouseOverListener>();
	private transient List<AppearanceChangeListener> appearanceListeners =
		new LinkedList<AppearanceChangeListener>();
	private transient final AppearanceChangeEvent appearanceEvent =
		new AppearanceChangeEvent(this);



	/** Creates a SceneComponent with all properties that are objects set to
	 * <code>null</code>. The three booleans controlling the participation in
	 * dragging are all set to <code>false</code>.*/
	
	public SceneComponent() {}
	
	public SceneComponent(SceneComponent s) {
		List<Point2D.Double> p = s.getPoints();
		for (Point2D.Double q : p) {
			points.add((Point2D.Double)q.clone());
		}
		List<Annotation> a = s.getAnnotations();
		for (Annotation b : a) {
			annotations.add(new Annotation(b));
		}
		List<Icon2D> i = s.getIcons();	
		for (Icon2D j : i) {
			icons.add(copyIcon2D(j));
		}
		shape = (Shape)copy(s.getShape());
		copyAttributesFrom(s);
	}
	
	public void copyAttributesFrom(SceneComponent s) {
		
		AffineTransform t = s.getTransform();
		transform = t==null ? null : (AffineTransform)t.clone();
		
		// use own copy method to copy interface types
		paint = copyPaint(s.getPaint());
		outlinePaint = copyPaint(s.getOutlinePaint());
		pointPaint = copyPaint(s.getPointPaint());
		annotationPaint = copyPaint(s.getAnnotationPaint());
		highlightPaint = copyPaint(s.getHighlightPaint());
		highlightOutlinePaint = copyPaint(s.getHighlightOutlinePaint());
		highlightPointPaint = copyPaint(s.getHighlightPointPaint());
		highlightAnnotationPaint = copyPaint(s.getHighlightAnnotationPaint());
		stroke = copyStroke(s.getStroke());;
		pointStroke = copyStroke(s.getPointStroke());
		pointShape = (Shape)copy(s.getPointShape());
		
		// copy immutable fields by reference
		annotationMargin = s.getAnnotationMargin();
		font = s.getFont();
		filled = s.getFilled();
		outlined = s.getOutlined();
		highlighted = s.getHighlighted();
		pointFilled = s.getPointFilled();
		pointOutlined = s.getPointOutlined();
		annotated = s.getAnnotated();
		areaDragEnabled = s.isAreaDragEnabled();
		outlineDragEnabled = s.isOutlineDragEnabled();
		pointDragEnabled = s.isPointDragEnabled();
		visible = s.isVisible();
		name = s.getName();
	}
	public Shape getShape() {
		return shape;
	}
	
	/** Sets the shape of this node. */
	
	public void setShape(Shape s) {
		shape = s;
	}
	
	public List<Point2D.Double> getPoints() {
		return points;
	}
	
	public List<Annotation> getAnnotations() {
		return annotations;
	}
	
	public Paint getPaint() {
		return paint;
	}
	
	/** Sets the paint to be used for filling the shape. A null value
	 * indicates that the paint should be inherited from the scene hierarchy. */
	
	public void setPaint(Paint p) {
		paint = p;
	}
	
	public Paint getOutlinePaint() {
		return outlinePaint;
	}
	
	/** Sets the paint to be used for drawing the outline of the shape or
	 * of the point set. A null value indicates that the outline paint should
	 * be inherited from the scene hierarchy. */
	
	public void setOutlinePaint(Paint p) {
		outlinePaint = p;
	}
	public Paint getPointPaint() {
		return pointPaint;
	}
	
	/** Sets the paint to be used for filling the points. A null value
	 * indicates that the point paint should be inherited from the scene
	 * hierarchy. */
	
	public void setPointPaint(Paint p) {
		pointPaint = p;
	}
	
	public Paint getAnnotationPaint() {
		return annotationPaint;
	}
	
	/** Sets the paint to be used for drawing the text of the annotation.
	 * A null value indicates that the annotation paint should be inherited
	 * from the scene hierarchy. */
	
	public void setAnnotationPaint(Paint p) {
		annotationPaint = p;
	}

	public Stroke getStroke() {
		return stroke;
	}
	
	/** Sets the stroke to be used for drawing the outline of the shape. A
	 * null value indicates that the stroke should be inherited from the scene
	 * hierarchy. */
	
	public void setStroke(Stroke s) {
			stroke=s;
	}
	public Stroke getPointStroke() {
		return pointStroke;
	}
	
	/** Sets the stroke to be used for drawing the outline of the points of
	 * this node. A null value indicates that the point stroke should be
	 * inherited from the scene hierarchy. */
	
	public void setPointStroke(Stroke s) {
		pointStroke = s;
	}
	
	public Shape getPointShape() {
		return pointShape;
	}
	
	/** Sets the stamp to be used for creating a shape for each point in the
	 * point set. A null value indicates that the stroke should be inherited
	 * from the scene hierarchy. */
	
	public void setPointShape(Shape s) {
		pointShape = s;
	}
	
	public Integer getAnnotationMargin() {
		return annotationMargin;
	}
	
	/** Sets the amount (in screen space) by which the text of the annotation
	 * is shifted up and to the right relative to points inducated by the
	 * annotation. */
	
	public void setAnnotationMargin(Integer a) {
		annotationMargin = a;
	}
	
	public Font getFont() {
		return font;
	}
	
	/** Sets the stroke to be used for drawing the outline of the points of
	 * this node. A null value indicates that the point stroke should be
	 * inherited from the scene hierarchy. */
	
	public void setFont(Font f) {
		font = f;
	}
	
	public AffineTransform getTransform() {
		return transform;
	}
	
	/** Sets the affine transformation of this node (mapping coordinates in
	 * the parent node to this node's own coordinates). */
	
	public void setTransform(AffineTransform t) {
		transform = t;
	}
	
	public Boolean getFilled() {
		return filled;
	}
	
	/** Sets the Boolean that tells whether the shape of this node should
	 * be filled. A null value indicates that thhis property should be inherited
	 * from the scene hierarchy. */
	
	public void setFilled(Boolean b) {
		filled = b;
	}
	
	public Boolean getOutlined() {
		return outlined;
	}
	
	/** Sets the Boolean that tells whether the outline of this node's shape
	 * should be drawn. A null value indicates that this property should be
	 * inherited from the scene hierarchy. */
	
	public void setOutlined(Boolean b) {
		outlined = b;
	}
	
	public Boolean getPointFilled() {
		return pointFilled;
	}
	
	/** Sets the Boolean that tells whether the points of this node should
	 * be filled. A null value indicates that this property should be inherited
	 * from the scene hierarchy. */
	
	public void setPointFilled(Boolean b) {
		pointFilled= b;
	}
	
	public Boolean getAnnotated() {
		return annotated;
	}
	
	/** Sets the Boolean that tells whether the textual annotation of this
	 * node should be drawn. A null value indicates that this property should
	 * be inherited from the scene hierarchy. */
	
	public  void setAnnotated(Boolean b) {
		annotated = b;
	}

	public Boolean getPointOutlined() {
		return pointOutlined;
	}
	
	/** Sets the Boolean that tells whether the outline of the points of this
	 * node should be drawn. A null value indicates that this property should
	 * be inherited from the scene hierarchy. */
	
	public void setPointOutlined(Boolean b) {
		pointOutlined = b;
	}
	
	public boolean isAreaDragEnabled() {
		return areaDragEnabled;
	}
	
	/** Determine whether the shape of this node takes part in picking. */
	
	public void setAreaDragEnabled(boolean b) {
		areaDragEnabled = b;
	}
	
	public boolean isOutlineDragEnabled() {
		return outlineDragEnabled;
	}
	
	/** Determine whether the outline of this node's shape takes part in
	 * picking. */
	
	public void setOutlineDragEnabled(boolean b) {
		outlineDragEnabled = b;
	}
	
	public boolean isPointDragEnabled() {
		return pointDragEnabled;
	}
	/** Tells whether the points of this node (or rather the shapes created
	 * by this node's stamp) take part in picking. */
	
	public void setPointDragEnabled(boolean b) {
		pointDragEnabled = b;
	}
	
	/** Adds a child node. */
	
	public void addChild(SceneComponent s) {
		children.add(s);
		s.addAppearanceChangeListener(this);
	}
	
	public SceneComponent getChild(int index) {
		return children.get(index);
	}
	
	public int getChildCount() {
		return children.size();
	}

	public void removeAllChildren() {
		children.clear();
	}
	
	/** Removes a child node. */
	
	public void removeChild(SceneComponent node) {
		children.remove(node);
	}
	
	public void dragStart(TransformedMouseEvent e) {
		for (DragListener l : dragListeners) {
			l.dragStart(e);
		}
	}
	
	public void drag(TransformedMouseEvent e) {
		for (DragListener l : dragListeners) {
			l.drag(e);
		}
	}
	
	public void dragEnd(TransformedMouseEvent e) {
		for (DragListener l : dragListeners) {
			l.dragEnd(e);
		}
	}

	/** Adds <code>l</code> to the list of listeners that are notified when
	 * this node is dragged. */
	
	public void addDragListener(DragListener l) {
		dragListeners.add(l);
	}

	/** Remove <code>l</code> from the list of listeners that are notified when
	 * this node is component. */
	
	public void removeDragListener(DragListener l) {
		dragListeners.remove(l);
	}
	
	/** Adds <code>l</code> to the list of listeners that are notified when
	 * this node is over this component. */
	
	public void addMouseOverListener(MouseOverListener l) {
		mouseOverListeners.add(l);
	}

	/** Remove <code>l</code> from the list of listeners that are notified when
	 * this mouse is over this component. */
	
	public void removeMouseOverListener(MouseOverListener l) {
		mouseOverListeners.remove(l);
	}
	
	
	/** Notifies all registered AppearanceChangeListeners that the appearance
	 * of this node has changed (meaning that a repaint is needed). */
	
	public void fireAppearanceChange() {
		for (AppearanceChangeListener l : appearanceListeners) {
			l.appearanceChange(appearanceEvent);
		}
	}

	public void addAppearanceChangeListener(AppearanceChangeListener l) {
		appearanceListeners.add(l);
	}

	public void removeAppearanceChangeListener(AppearanceChangeListener l) {
		appearanceListeners.remove(l);
	}
	
	public void appearanceChange(AppearanceChangeEvent e) {
		fireAppearanceChange();
	}
	
	public boolean isVisible() {
		return visible;
	}
	
	public void setVisible(boolean b) {
		visible = b;
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}
	
	private Paint copyPaint(Paint p) {
		if (p instanceof Color) {
			Color q = (Color)p;
			return new Color(q.getRed(), q.getGreen(),q.getBlue(),q.getAlpha());
		} else {
			return (Paint)copy(p);
		}
	}
	
	private Stroke copyStroke(Stroke stroke) {
		if (stroke instanceof BasicStroke) {
			BasicStroke s = (BasicStroke)stroke;
			return new BasicStroke(
					s.getLineWidth(),
					s.getEndCap(),
					s.getLineJoin(),
					s.getMiterLimit(),
					s.getDashArray(),
					s.getDashPhase()
			);
		} else {
			return null;
		}
	}
	
	private Icon2D copyIcon2D(Icon2D i) {
		if (i == null) return null;
		Icon icon = i.getIcon();
		Icon newIcon = null;
		if (icon != null) {
			if (icon instanceof ImageIcon) {
				ImageIcon imageIcon = (ImageIcon)icon;
				BufferedImage image = new BufferedImage(
						icon.getIconWidth(),
						icon.getIconHeight(),
						BufferedImage.TYPE_INT_ARGB
				);
				Graphics2D g = image.createGraphics();
				g.drawImage(imageIcon.getImage(),0,0,null);
			}
		} else {
			newIcon = (Icon)copy(icon);
		}
		return new Icon2D(newIcon, i.getX(), i.getY(), i.getAnchor());
	}
	
	private Object copy(Object o) {
		final Object[] noArguments = new Object[0];
		Object result = null;
		if (o!=null) {
			Expression exp;
			try {
				if (o instanceof Cloneable) {
					exp = new Expression(o, "clone", noArguments);
					result = exp.getValue();
				} else {
					// copy contructor
					exp = new Expression(o.getClass(), "new", new Object[]{o});
					result = exp.getValue();
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		return result;
	}

	public boolean isMouseOverEnabled() {
		return mouseOverEnabled;
	}

	public void setMouseOverEnabled(boolean areaMouseOverEnabled) {
		this.mouseOverEnabled = areaMouseOverEnabled;
	}
	
	public void mouseEnter(TransformedMouseEvent e) {
		int n = mouseOverListeners.size();
		for (int i=0; i<n; i++) {
			mouseOverListeners.get(i).mouseEnter(e);
		}
	}
	
	public void mouseLeave(TransformedMouseEvent e) {
		int n = mouseOverListeners.size();
		for (int i=0; i<n; i++) {
			mouseOverListeners.get(i).mouseLeave(e);
		}
	}

	public List<Icon2D> getIcons() {
		return icons;
	}

	public boolean isIconDragEnabled() {
		return iconDragEnabled;
	}

	public void setIconDragEnabled(boolean iconDragEnabled) {
		this.iconDragEnabled = iconDragEnabled;
	}

	public Boolean getHighlighted() {
		return highlighted;
	}

	public void setHighlighted(Boolean highlighted) {
		this.highlighted = highlighted;
	}

	public Paint getHighlightOutlinePaint() {
		return highlightOutlinePaint;
	}

	public void setHighlightOutlinePaint(Paint highlightOutlinePaint) {
		this.highlightOutlinePaint = highlightOutlinePaint;
	}

	public Paint getHighlightPaint() {
		return highlightPaint;
	}

	public void setHighlightPaint(Paint highlightPaint) {
		this.highlightPaint = highlightPaint;
	}

	public Paint getHighlightPointPaint() {
		return highlightPointPaint;
	}

	public void setHighlightPointPaint(Paint highlightPointPaint) {
		this.highlightPointPaint = highlightPointPaint;
	}

	public Paint getHighlightAnnotationPaint() {
		return highlightAnnotationPaint;
	}

	public void setHighlightAnnotationPaint(Paint highlightAnnotationPaint) {
		this.highlightAnnotationPaint = highlightAnnotationPaint;
	}

	public boolean isAnnotationDragEnabled() {
		return annotationDragEnabled;
	}

	public void setAnnotationDragEnabled(boolean annotationDragEnabled) {
		this.annotationDragEnabled = annotationDragEnabled;
	}
}
