/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.java2dx.modelling;


import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.GeneralPath;
import java.awt.geom.Ellipse2D;
import java.io.IOException;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import de.jtem.java2d.SceneComponent;
import de.jtem.java2dx.Line2DDouble;

public class DraggableLine2D extends ModellingTool2D {
	private static final long serialVersionUID = -5623989232010175729L;
	
	public final static int NONE = 0;
	public final static int ON_TIP = 1;
	public final static int ON_BOTH_ENDS = 2;
	
	private Line2DDouble line;
	private int arrowHeadCount =2;
	private boolean circlesInsteadOfArrows = false;
	private double arrowHeadRadius = 5;
	private double arrowHeadAspectRatio = .7;
	private transient DraggablePoint2D p1;
	private transient DraggablePoint2D p2;
	private transient AffineTransform headTransform;
	private transient AffineTransform tailTransform;
	private transient Shape headShape;
	private transient Shape tailShape;
	private transient Shape arrowShape;
	private Shape circleShape = new Ellipse2D.Double(-1,-1,2,2);
	
	private Line2DDouble storedLine = new Line2DDouble();
	private boolean storedCirclesInsteadOfArrows;
	private int storedArrowHeadCount;
	private double storedArrowHeadRadius;
	private double storedArrowHeadAspectRatio;
	
	private SceneComponent controlScene;
	private transient SceneComponent tail;
	
	public DraggableLine2D() {
		this(new Line2DDouble());
	}
	
	public DraggableLine2D(DraggableLine2D p, Line2DDouble q) {
		super(p);
		q.setLine(p.getModel());
		line = q;
		getViewScene().copyAttributesFrom(p.getViewScene());
		setUp();
	}
	
	/** makes a copy of the model of p */
	
	public DraggableLine2D(DraggableLine2D p) {
		this(p,new Line2DDouble(p.getModel()));
	}

	/** p is not copied */
	
	public DraggableLine2D(Line2DDouble p) {
		line = p;
		setUp();
	}
	
	private void setUp() {
		getViewScene().setShape(line);
		controlScene = new SceneComponent();
		p1 = new DraggablePoint2D(line.getP1());
		p2 = new DraggablePoint2D(line.getP2());
		
		// arrow heads
		tail = new SceneComponent();
		tail.getPoints().add(p1.getModel());
		//for serialization
		getViewScene().getPoints().clear();
		getViewScene().getPoints().add(p2.getModel());
		getViewScene().addChild(tail);
		headTransform = new AffineTransform();
		tailTransform = new AffineTransform();
		updateTransforms();
		updateArrowShape();
		
		controlScene.addChild(p1.getViewScene());
		controlScene.addChild(p2.getViewScene());
		p1.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				getModel().setLine(p1.getModel(), p2.getModel());
				fireModelChange();
				updateFromModel();
			}
		});
		p2.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				getModel().setLine(p1.getModel(), p2.getModel());
				fireModelChange();
				updateFromModel();
			}
		});
		
		getViewScene().addDragListener(new TranslationTool(this));
		storeModelState();
	}
	
	public Line2DDouble getModel() {
		return line;
	}

	public SceneComponent getControlScene() {
		return controlScene;
	}

	public void translate(double x, double y) {
		line.setLine(line.x1 + x, line.y1 + y, line.x2 + x, line.y2 + y);
	}

	public void storeModelState() {
		storedLine.setLine(getModel());
		storedArrowHeadAspectRatio = arrowHeadAspectRatio;
		storedArrowHeadCount = arrowHeadCount;
		storedCirclesInsteadOfArrows = circlesInsteadOfArrows;
	}

	public void recallModelState() {
		getModel().setLine(storedLine);
		arrowHeadAspectRatio = storedArrowHeadAspectRatio;
		arrowHeadCount = storedArrowHeadCount;
		arrowHeadRadius = storedArrowHeadRadius;
		circlesInsteadOfArrows = storedCirclesInsteadOfArrows;
		fireModelChange();
		updateFromModel();
	}
	
	public void updateFromModel() {
		p1.getModel().setLocation(line.x1, line.y1);
		p1.updateFromModel();
		p2.getModel().setLocation(line.x2, line.y2);
		p2.updateFromModel();
		updateTransforms();
	}

	public boolean isCirclesInsteadOfArrows() {
		return circlesInsteadOfArrows;
	}

	public void setCirclesInsteadOfArrows(boolean b) {
		if (b != circlesInsteadOfArrows) {
			circlesInsteadOfArrows = b;
			updatePointShapes();
			getViewScene().fireAppearanceChange();
		}
	}

	public double getArrowHeadAspectRatio() {
		return arrowHeadAspectRatio;
	}

	public void setArrowHeadAspectRatio(double a) {
		if (a != arrowHeadAspectRatio) {
			arrowHeadAspectRatio = a;
			updateArrowShape();
			updatePointShapes();
			getViewScene().fireAppearanceChange();
		}
	}

	public int getArrowHeadCount() {
		return arrowHeadCount;
	}

	public void setArrowHeadCount(int n) {
		// we have at most two heads
		n = Math.max(0,n);
		n = Math.min(2,n);
		if (n!= arrowHeadCount) {
			arrowHeadCount = n;
			updatePointShapes();
		}
		getViewScene().fireAppearanceChange();
	}

	public double getArrowHeadRadius() {
		return arrowHeadRadius;
	}

	public void setArrowHeadRadius(double r) {
		if (r!= arrowHeadRadius) {
			arrowHeadRadius = r;
			updateTransforms();
			getViewScene().fireAppearanceChange();
		}
	}
	
	public DraggableLine2D clone() {
		return new DraggableLine2D(this);
	}
	
	private void updatePointShapes() {
		headShape = headTransform.createTransformedShape(
				circlesInsteadOfArrows ? circleShape : arrowShape
		);
		tailShape = tailTransform.createTransformedShape(
				circlesInsteadOfArrows ? circleShape : arrowShape
		);
		tail.setPointShape(tailShape);
		tail.setPointFilled(Boolean.valueOf(arrowHeadCount > 1));
		tail.setPointOutlined(Boolean.valueOf(arrowHeadCount > 1));
		getViewScene().setPointShape(headShape);
		getViewScene().setPointFilled(Boolean.valueOf(arrowHeadCount > 0));
		getViewScene().setPointOutlined(Boolean.valueOf(arrowHeadCount > 0));
	}
	
	private void updateTransforms() {
		double a = line.x2 - line.x1;
		double b = line.y2 - line.y1;
		double norm = Math.sqrt(a*a + b*b);
		if (norm != 0) {
			double factor = arrowHeadRadius/norm;
			a *= factor;
			b *= factor;
			headTransform.setTransform(a, -b, b, a, 0, 0);
			tailTransform.setTransform(-a, b, -b, -a, 0, 0);
			updatePointShapes();
		}
	}
	
	private void updateArrowShape() {
		float h = (float) arrowHeadAspectRatio;
		GeneralPath path = new GeneralPath();
		path.moveTo(0,0);
		path.lineTo(-2, h);
		path.lineTo(-2, -h);
		path.closePath();
		arrowShape = path;
		updatePointShapes();
	}
	
	private void readObject(java.io.ObjectInputStream stream)
	throws IOException, ClassNotFoundException {
		stream.defaultReadObject();
		setUp();
	}
}
