/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.java2dx.modelling;


import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.awt.geom.Point2D;
import java.io.IOException;
import java.util.List;

import javax.swing.AbstractAction;

import de.jtem.java2d.TransformedMouseEvent;
import de.jtem.java2d.DragListener;
import de.jtem.java2d.SceneComponent;
import de.jtem.java2dx.Point2DDouble;
import de.jtem.java2dx.Point2DList;
import de.jtem.modelling.ActionTree;

public class DraggablePoint2DList extends ModellingTool2D {

	private static final long serialVersionUID = -7254181073831753423L;
	public static final int CIRCULAR = 0;
	public static final int LINEAR = 1;
	public static final int RANDOM = 2;
	
	private Point2DList pointList;
	private int defaultPointCount = 1;
	private int defaultArrangement = LINEAR;
	private double defaultTranslationX;
	private double defaultTranslationY;
	private double defaultRadius;
	private  Point2DList storedPoints = new Point2DList();
	
	private transient int button;
	private transient double lastX;
	private transient double lastY;
	private transient double centerX;
	private transient double centerY;
	private transient double stretchRotationX = 1;
	private transient double stretchRotationY = 0;
	private transient DragListener transformationTool;
	
	public DraggablePoint2DList() {
		this(1,DraggablePoint2DList.LINEAR,0,0,1);
	}
	
	public DraggablePoint2DList(
			int defaultPointCount,
			int defaultArrangement,
			double defaultTranslationX,
			double defaultTranslationY,
			double defaultRadius
	) {
		this.defaultArrangement = defaultArrangement;
		this.defaultTranslationX = defaultTranslationX;
		this.defaultTranslationY = defaultTranslationY;
		this.defaultRadius = defaultRadius;
		this.defaultPointCount = defaultPointCount;
		pointList = new Point2DList();
		setControlPointsToDefault();
		storeModelState();
		setUp();
	}
	
	/**
	 * p is not copied.
	 * 
	 * @param p
	 * @param arrangementType
	 * @param positionX
	 * @param positionY
	 * @param radius
	 */
	public DraggablePoint2DList(
			Point2DList p,
			int defaultArrangement,
			double defaultTranslationX,
			double defaultTranslationY,
			double defaultRadius
	) {
		pointList = p;
		this.defaultArrangement = defaultArrangement;
		this.defaultTranslationX = defaultTranslationX;
		this.defaultTranslationY = defaultTranslationY;
		this.defaultRadius = defaultRadius;
		storeModelState();
		setUp();
	}
	

	/**
	 * Creates a copy of p with a new PointList that is a copy of the one of p.
	 * 
	 * @param p the DraggablePointSet to be copied
	 */
	
	public DraggablePoint2DList(DraggablePoint2DList p) {
		super(p);
		defaultArrangement = p.getDefaultArrangement();
		defaultTranslationX = p.getDefaultTranslationX();
		defaultTranslationY = p.getDefaultTranslationY();
		defaultRadius = p.getDefaultRadius();
		defaultPointCount = p.getDefaultPointCount();
		pointList = new Point2DList(p.getModel());
		storeModelState();
		setUp();
		getViewScene().copyAttributesFrom(p.getViewScene());
	}
	
	public void reset() {
		setControlPointsToDefault();
		updateModelFromController();
		fireModelChange();
		updateViewFromModel();
	}
	
	public int getDefaultArrangement() {
		return defaultArrangement;
	}

	public void setDefaultArrangement(int defaultArrangement) {
		this.defaultArrangement = defaultArrangement;
	}

	public double getDefaultRadius() {
		return defaultRadius;
	}

	public void setDefaultRadius(double defaultRadius) {
		this.defaultRadius = defaultRadius;
	}

	public double getDefaultTranslationX() {
		return defaultTranslationX;
	}

	public void setDefaultTranslationX(double defaultTranslationX) {
		this.defaultTranslationX = defaultTranslationX;
	}

	public double getDefaultTranslationY() {
		return defaultTranslationY;
	}

	public void setDefaultTranslationY(double defaultTranslationY) {
		this.defaultTranslationY = defaultTranslationY;
	}

	public Point2DList getModel() {
		return pointList;
	}

	public SceneComponent getControlScene() {
		return null;
	}

	public void translate(double x, double y) {
		int n = pointList.size();
		for (int j=0; j<n; j++) {
			double pX = pointList.get(j).x;
			double pY = pointList.get(j).y;
			pX += x;
			pY += y;
			pointList.get(j).setLocation(pX,pY);
		}
		defaultTranslationX += x;
		defaultTranslationY += y;
		fireModelChange();
		updateViewFromModel();
	}
	
	public int getDefaultPointCount() {
		return defaultPointCount;
	}
	
	public void setDefaultPointCount(int defaultPointCount) {
		this.defaultPointCount = defaultPointCount;
	}
	
	public DraggablePoint2DList clone() {
		return new DraggablePoint2DList(this);
	}
	
	private void updateModelFromController() {}
	
	protected void updateFromModel() {
		updateViewFromModel();
		getViewScene().fireAppearanceChange();
	}
	
	protected void updateViewFromModel() {
		List<Point2D.Double> points = getViewScene().getPoints();
		points.clear();
		points.addAll(pointList);
	}
	
	private void setUp() {
		transformationTool = new DragListener() {
			
			public void dragStart(TransformedMouseEvent e) {
				MouseEvent m = e.getMouseEvent();
				if (e.getIndexOfHitPart()<0 || m.isShiftDown()) {
					button = e.getButton();
					lastX = e.getX();
					lastY = e.getY();
					if (button==2) updateCenter();
				}
			}
			
			public void drag(TransformedMouseEvent e) {
				MouseEvent m = e.getMouseEvent();
				if (e.getIndexOfHitPart()<0 || m.isShiftDown()) {
					double x = e.getX();
					double y = e.getY();
					
					switch(button) {
					
					case 1 : {
						translate(x-lastX, y-lastY);
						break;
					}
					
					case 2 : {
						double aX = x-centerX;
						double aY = y-centerY;
						double bX = lastX-centerX;
						double bY = lastY-centerY;
						
						double normSquared = bX*bX + bY*bY;
						double cX = (aX*bX + aY*bY)/normSquared;
						double cY = (aY*bX - aX*bY)/normSquared;
						stretchRotate(cX, cY);
					}
					}
					lastX=x;
					lastY=y;
					fireModelChange();
					updateViewFromModel();
				}
			}
			
			public void dragEnd(TransformedMouseEvent e) {}
		};
		getViewScene().addDragListener(transformationTool);
		getViewScene().addDragListener(new DragListener() {
			int index;
			public void dragStart(TransformedMouseEvent e) {
				index = e.getIndexOfHitPart();
				lastX = e.getX();
				lastY = e.getY();
			}
			
			public void drag(TransformedMouseEvent e) {
				double x = e.getX();
				double y = e.getY();
				double translationX = x-lastX;
				double translationY = y-lastY;
				double pX = pointList.get(index).x;
				double pY = pointList.get(index).y;
				pX += translationX;
				pY += translationY;
				pointList.get(index).setLocation(pX,pY);
				lastX=x;
				lastY=y;
				updateModelFromController();
				fireModelChange();
				updateViewFromModel();
				getViewScene().fireAppearanceChange();
			}
			
			public void dragEnd(TransformedMouseEvent e) {}
		});
		addActionTree(new ActionTree(
				new AbstractAction("reset") {
					private static final long serialVersionUID = 1400384559035275571L;
					
					public void actionPerformed(ActionEvent e) {
						reset();
					}
				}));
		SceneComponent view = getViewScene();
		view.setPointDragEnabled(true);
		view.getPoints().clear();
		view.getPoints().addAll(pointList);
	}
	
	public DragListener getTransformationTool() {
		return transformationTool;
	}
	
	private void stretchRotate(double cX, double cY) {
		double newStretchRotationX = stretchRotationX * cX - stretchRotationY * cY;
		stretchRotationY = stretchRotationX * cY + stretchRotationY * cX;
		stretchRotationX = newStretchRotationX;
		int n = pointList.size();
		for (int j=0; j<n; j++) {
			double dX = pointList.get(j).x - centerX;
			double dY = pointList.get(j).y - centerY;
			pointList.get(j).setLocation(
					centerX + cX*dX - cY*dY,
					centerY + cX*dY + cY*dX
			);
		}
		updateModelFromController();
		fireModelChange();
		updateViewFromModel();
	}
	
	
	private void setControlPointsToDefault() {
		pointList.clear();
		int n = defaultPointCount;
    	double x;
    	double y=0;
    	if (defaultArrangement == LINEAR) {
    		final double dt = 1/(double)(n-1);
    		for (int j=0; j<n; j++) {
    			double t = j*dt;
    			x = defaultTranslationX + defaultRadius * (-1+2*t);
    			pointList.add(new Point2DDouble(x,y));
    		}
    	}
    	if (defaultArrangement == CIRCULAR) {
    		final double dt = 2*Math.PI/(double)n;
    		for (int j=0; j<n; j++) {
    			double t = j*dt;
    			x = defaultTranslationX + defaultRadius*Math.cos(t);
    			y = defaultTranslationY + defaultRadius*Math.sin(t);
    			pointList.add(new Point2DDouble(x,y));
    		}
		}
    	if (defaultArrangement == RANDOM) {
    		final double r = defaultRadius * defaultRadius;
    		int j = 0;
    		while (j<n) {
    			x = -1+2*Math.random();
    			y = -1+2*Math.random();
    			if (x*x+y*y < r) {
    				pointList.add(
    						new Point2DDouble(
    								defaultTranslationX + defaultRadius*x,
    								defaultTranslationY + defaultRadius*y
    						)
    				);
    				j++;
    			}
    		}
    	}
	}
	
	private void updateCenter() {
		double minX = Double.MAX_VALUE;
		double maxX = -Double.MAX_VALUE;
		double minY = Double.MAX_VALUE;
		double maxY = -Double.MAX_VALUE;
		int n = pointList.size();
		for (int j=0; j<n; j++) {
			double x = pointList.get(j).x;
			double y = pointList.get(j).y;
			if (x < minX) minX = x;
			if (x > maxX) maxX = x;
			if (y < minY) minY = y;
			if (y > maxY) maxY = y;
		}
		centerX = .5*(minX + maxX);
		centerY = .5*(minY + maxY);
	}

	public void storeModelState() {
		storedPoints.assign(pointList);
	}

	public void recallModelState() {
		pointList.assign(storedPoints);
		fireModelChange();
		updateViewFromModel();
	}
	
	private void readObject(java.io.ObjectInputStream stream)
	throws IOException, ClassNotFoundException {
		stream.defaultReadObject();
		setUp();
	}
}