/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.java2dx.modelling;


import java.awt.event.MouseEvent;
import java.io.IOException;
import java.io.Serializable;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

import javax.swing.JPopupMenu;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import de.jtem.java2d.DragListener;
import de.jtem.java2d.MouseOverListener;
import de.jtem.java2d.SceneComponent;
import de.jtem.java2d.TransformedMouseEvent;
import de.jtem.modelling.ActionTree;
import de.jtem.modelling.ModellingTool;

public abstract class ModellingTool2D implements ChangeListener, ModellingTool, Serializable {
	private String name;
	private String modelType;
	private SceneComponent viewScene = new SceneComponent();
	private transient JPopupMenu menu = new JPopupMenu();
	private transient double positionX;
	private transient double positionY;
	private transient List<ChangeListener> changeListeners=new CopyOnWriteArrayList<ChangeListener>();
	private transient DragListener menuPopup;
	public static class TranslationTool implements DragListener {
		ModellingTool2D modellingTool;
		TranslationTool(ModellingTool2D m) {
			modellingTool = m;
		}
		double lastX;
		double lastY;
		public void dragStart(TransformedMouseEvent e) {
			lastX = e.getX();
			lastY = e.getY();
		}
		public void drag(TransformedMouseEvent e) {
			double newX = e.getX();
			double newY = e.getY();
			modellingTool.translate(newX - lastX, newY - lastY);
			lastX = newX;
			lastY = newY;
			modellingTool.fireModelChange();
			modellingTool.updateFromModel();
		}
		public void dragEnd(TransformedMouseEvent e) {}
	};
	
	protected ModellingTool2D() {
		setUp();
	}
	
	protected ModellingTool2D(ModellingTool2D m) {
		positionX = m.positionX;
		positionY = m.positionY;
		modelType = m.modelType;
		setUp();
	}
	
	/**
	 * Overrides the method from the <code>ModellingTool</code> interface with
	 * an appropriate return type.
	 */
	public SceneComponent getViewScene() {
		return viewScene;
	}
	
	/**
	 * Overrides the method from the <code>ModellingTool</code> interface with
	 * an appropriate return type.
	 */
	public abstract SceneComponent getControlScene();
	
	public abstract void translate(double x, double y);

	/**
	 * Respond to a change of the model state.
	 */
	protected abstract void updateFromModel();
	
	public void moveTo(double[] v) {
		translate(v[0]-positionX, v[1]-positionY);
		updateFromModel();
		positionX = v[0];
		positionY = v[1];
	}
	
	public String getName() {
		return name;
	}
	
	public void setName(String s) {
		name = s;
	}
	
	public String getModelType() {
		return modelType;
	}
	
	public void setModelType(String s) {
		modelType = s;
	}
	
	public void stateChanged(ChangeEvent e) {
		if (e.getSource() == getModel()) {
			updateFromModel();
		}
	}
	
	public void addActionTree(ActionTree t) {
			menu.add(t.asMenuItem());
	}

	public void fireModelChange() {
		for (ChangeListener changeListener : changeListeners) {
			changeListener.stateChanged(new ChangeEvent(this));
		}
	}
	
	public void addChangeListener(ChangeListener listener) {
		changeListeners.add(listener);
	}
	
	public void removeChangeListener(ChangeListener listener) {
		changeListeners.remove(listener);
	}

	public JPopupMenu getMenu() {
		return menu;
	}
	public DragListener getMenuDragListener() {
		return menuPopup;
	}
	
	public void setMenuDragListener(DragListener l) {
		viewScene.removeDragListener(menuPopup);
		viewScene.addDragListener(l);
		menuPopup = l;
	}
	
	public abstract ModellingTool2D clone();
	
	private void setUp() {
		menuPopup = new DragListener() {

			public void dragStart(TransformedMouseEvent e) {
				if (e.getMouseEvent().getButton()==3) {
	    			MouseEvent m = e.getMouseEvent();
	    			positionX = e.getX();
	    			positionY = e.getY();
	    			menu.show(m.getComponent(), m.getX(), m.getY());
	    		}
			}
			public void drag(TransformedMouseEvent e) {}
			public void dragEnd(TransformedMouseEvent e) {}
		};
		viewScene.addMouseOverListener(
				new MouseOverListener() {
					public void mouseEnter(TransformedMouseEvent e) {
						viewScene.setHighlighted(Boolean.TRUE);
						viewScene.fireAppearanceChange();
					}
					public void mouseLeave(TransformedMouseEvent e) {
						viewScene.setHighlighted(null);
						viewScene.fireAppearanceChange();
					}
				});
		viewScene.setMouseOverEnabled(true);
		viewScene.setAreaDragEnabled(true);
		viewScene.setPointDragEnabled(true);
		viewScene.setOutlineDragEnabled(true);
		viewScene.addDragListener(menuPopup);
	}
	
	private void readObject(java.io.ObjectInputStream stream)
	throws IOException, ClassNotFoundException {
		stream.defaultReadObject();
		setUp();		
	}

}
