/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.java2dx.modelling;


import java.awt.Color;
import java.awt.Font;
import java.awt.Paint;
import java.awt.Stroke;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Rectangle2D;
import java.io.File;
import java.security.AccessControlException;
import java.util.HashMap;
import java.util.Set;

import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JPopupMenu;
import javax.swing.event.ChangeListener;
import javax.swing.filechooser.FileSystemView;

import de.jtem.java2d.SceneComponent;
import de.jtem.java2dx.beans.Viewer2DWithInspector;
import de.jtem.modelling.ActionTree;
import de.jtem.modelling.Inspector;
import de.jtem.modelling.Modeller;
import de.jtem.modelling.ModellingViewer;
import de.jtem.modelling.PersistencePeer;

public class SimpleModeller2D implements ModellingViewer, Inspector, PersistencePeer {

	private static final long serialVersionUID = 933354580637372378L;
    
	private Modeller modeller;
	private JFileChooser chooser;
	private SceneComponent backRoot;
	private Viewer2DWithInspector viewer;
	
	public SimpleModeller2D() {
		this(new Viewer2DWithInspector(new Rectangle2D.Double(-3,-3,6,6)));
	}
	
    public SimpleModeller2D(Viewer2DWithInspector viewer) {
    	this.viewer = viewer;
    	backRoot = new SceneComponent();
    	viewer.getRoot().addChild(backRoot);
    	
    	// control scene
    	SceneComponent controlScene = new SceneComponent();
    	controlScene.setName("control scene");
        controlScene.setPointShape(new Ellipse2D.Double(-8,-8,16,16));
        controlScene.setPointPaint(new Color(255,255,0,80));
        controlScene.setOutlinePaint(Color.red);
        
        // view scene;
        SceneComponent viewScene = new SceneComponent();
    	viewScene.setName("view scene");
    	
        //TODO get lastDir from preferences
    	try {
    	       File lastDir = null;
    	        FileSystemView view = FileSystemView.getFileSystemView();
    			chooser = new JFileChooser((lastDir == null || !lastDir.exists()) ? view
    					.getHomeDirectory() : lastDir, view);
    	        
    	    		
    	}catch(AccessControlException e)	{
    		e.printStackTrace();
    	}
        // modeller
    	modeller = new Modeller(
        		null,
    			this,
    			this,
    			this,
    			controlScene,
    			viewScene
        );
    	
    	modeller.setExcludedProperties(
    			viewScene,
    			Modeller2DUtility.getExpertProperties()
    	);
    	modeller.setExcludedProperties(
    			controlScene,
    			Modeller2DUtility.getExpertProperties()
    	);
    	
    	// add some objects to the inspect menu
		modeller.addObjectToInspect(
				getViewer(),
				"viewer"
		);
		modeller.setExcludedProperties(
				getViewer(),
				Modeller2DUtility.getExpertPropertiesForViewer2D()
		);
		modeller.addObjectToInspect(viewer.getGrid(),"grid");
		
		modeller.getFileMenu().add(viewer.getExportPDFAction());
		
		ActionTree viewerMenu = new ActionTree("inspect");
		viewerMenu.add(modeller.getInspectMenu());
		viewerMenu.add(viewer.getExportPDFAction());
		setActionTree(viewerMenu);
    }

	public void render() {
		viewer.repaint();
	}

	public void setActionTree(ActionTree t) {
		JPopupMenu menu = viewer.getMenu();
		menu.removeAll();
		int n = t.getItemCount();
		for (int j=0; j<n; j++) {
			menu.add(t.getItem(j).asMenuItem());
		}
	}

	public void addChildSceneToParentScene(Object child, Object parent) {
		((SceneComponent)parent).addChild((SceneComponent)child);
	}

	public void removeChildSceneFromParentScene(Object child, Object parent) {
		((SceneComponent)parent).removeChild((SceneComponent)child);
	}

	public void addScene(Object scene) {
		viewer.getRoot().addChild((SceneComponent) scene);
	}
	
	public void removeScene(Object scene) {
		viewer.getRoot().removeChild((SceneComponent) scene);
	}
	
	public double[] getLastMenuLocation() {
		return new double[]{
				viewer.getLastPopupLocationX(),
				viewer.getLastPopupLocationY()
		};
	}

	public void addChangeListener(ChangeListener l) {
		viewer.getInspectorPanel().addChangeListener(l);
	}

	public void removeChangeListener(ChangeListener l) {
		viewer.getInspectorPanel().removeChangeListener(l);
	}

	public Object getObject() {
		return viewer.getInspectorPanel().getObject();
	}
	
	public void storePersistentProperties(HashMap<String, Object> store) {
		
		// viewer
		store.put("gridEnabled", viewer.isGridEnabled());
		store.put("background", viewer.getBackground());
		store.put("snapToGridEnabled", viewer.isSnapToGridEnabled());
		store.put("snapToGridResolution", viewer.getSnapToGridResolution());
		store.put("viewport", viewer.getViewport());
		
		// grid
		store.put("annotated", viewer.getGrid().getAnnotated());
		store.put("annotationMargin", viewer.getGrid().getAnnotationMargin());
		store.put("coarseGridPaint", viewer.getGrid().getCoarseGridPaint());
		store.put("coarseGridStroke", viewer.getGrid().getCoarseGridStroke());
		store.put("font", viewer.getGrid().getFont());
		store.put("horizontalUnit", viewer.getGrid().getHorizontalUnit());
		store.put("outlinePaint", viewer.getGrid().getOutlinePaint());
		store.put("gridStroke", viewer.getGrid().getStroke());
		store.put("verticalUnit", viewer.getGrid().getVerticalUnit());
		store.put("maximumLineCount", viewer.getGrid().getMaximumLineCount());
	}
	
	public void restorePersistentProperties(HashMap<String, Object> store) {
		
		// viewer
		viewer.setGridEnabled((Boolean)store.get("gridEnabled"));
		viewer.setBackground((Color)store.get("background"));
		viewer.setSnapToGridEnabled((Boolean)store.get("snapToGridEnabled"));
		viewer.setSnapToGridResolution((Integer)store.get("snapToGridResolution"));
		viewer.encompass((Rectangle2D.Double)store.get("viewport"),0,0);
		
		// grid
		viewer.getGrid().setAnnotated((Boolean)store.get("annotated"));
		viewer.getGrid().setAnnotationMargin((Integer)store.get("annotationMargin"));
		viewer.getGrid().setCoarseGridPaint((Paint)store.get("coarseGridPaint"));
		viewer.getGrid().setFont((Font)store.get("font"));
		viewer.getGrid().setCoarseGridStroke((Stroke)store.get("coarseGridStroke"));
		viewer.getGrid().setHorizontalUnit((String)store.get("horizontalUnit"));
		viewer.getGrid().setVerticalUnit((String)store.get("verticalUnit"));
		viewer.getGrid().setOutlinePaint((Paint)store.get("outlinePaint"));
		viewer.getGrid().setStroke((Stroke)store.get("gridStroke"));
		viewer.getGrid().setMaximumLineCount((Integer)store.get("maximumLineCount"));
	}

	public Modeller getModeller() {
		return modeller;
	}

	public File openFileDialog() {
		chooser.showOpenDialog(null);
		return chooser.getSelectedFile();
	}

	public File saveFileDialog() {
		chooser.showSaveDialog(null);
		return chooser.getSelectedFile();
	}
	
	public SceneComponent getBackRoot() {
		return backRoot;
	}

	public Viewer2DWithInspector getViewer() {
		return viewer;
	}

	public void setViewer(Viewer2DWithInspector viewer) {
		this.viewer = viewer;
	}
	
    public static void main(String[] args) {
		JFrame frame = new JFrame("simple modeller");
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.setSize(600,600);
		SimpleModeller2D modeller = new SimpleModeller2D();
		modeller.getModeller().addTool(new DraggablePolygon2D(),null);
		frame.getContentPane().add(modeller.getViewer());
		frame.setVisible(true);
	}

	public void inspect(Object o, Set<String> excludedProperties) {
		getViewer().inspect(o, excludedProperties);
	}
}
