/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.java2dx.numerical;


import java.awt.BasicStroke;
import java.awt.Paint;
import java.awt.Stroke;
import java.awt.geom.GeneralPath;
import java.awt.geom.Rectangle2D;

import de.jtem.java2d.SceneComponent;
import de.jtem.numericalMethods.geometry.hyperSurface.MarchingSquares;

public class ContourPlot {
	
	Rectangle2D.Double rectangle = new Rectangle2D.Double(-1,-1,2,2);
	int horizontalGridLineCount = 100;
	int verticalGridLineCount = 100;
	RealFunctionOnRectangle function;
	double[][] func;
	double[] value;
	private String name = "contour plot";
	private double levelSpacing = .2;
	private int maximumLevelCount = 100;
	private double levelOffset;
	private SceneComponent view = new SceneComponent();
	
	public ContourPlot() {
		BasicStroke stroke = new BasicStroke(
				1,
				BasicStroke.CAP_ROUND,
				BasicStroke.JOIN_MITER
		);
		view.setOutlinePaint(
				new java.awt.Color(128,0,128,100)
		);
		view.setStroke(stroke);
		func = new double[horizontalGridLineCount][verticalGridLineCount];
		value = new double[horizontalGridLineCount*verticalGridLineCount];
	}
	
	void updateLevels() {
		if (function != null) {
			double minX = Math.max(
					rectangle.x,
					function.getLowerBoundOfXInterval()
			);
			double maxX = Math.min(
					rectangle.x+rectangle.width,
					function.getUpperBoundOfXInterval()
			);
			double minY = Math.max(
					rectangle.y,
					function.getLowerBoundOfYInterval()
			);
			double maxY = Math.min(
					rectangle.y+rectangle.height,
					function.getUpperBoundOfYInterval()
			);
			view.removeAllChildren();
			if (function == null || rectangle == null) return;
			double dx = (maxX-minX)/(horizontalGridLineCount-1);
			double dy = (maxY-minY)/(verticalGridLineCount-1);
			
			double min = Double.MAX_VALUE;
			double max = Double.MIN_VALUE;
			for (int i=0; i<horizontalGridLineCount; i++) {
				for (int j=0; j<verticalGridLineCount; j++) {
					double z = function.value(minX+i*dx, minY+j*dy);
					func[i][j]=z;
					if (z>max) {
						max = z;
					}
					if (z<min) {
						min = z;
					}
				}
			}
			
			// check if we have too many levels
			int numLevels = (int)((max-min)/levelSpacing);
			
			if (numLevels > maximumLevelCount) {
				for (int i=0; i<verticalGridLineCount; i++) {
					for (int j=0; j<horizontalGridLineCount; j++) {
						value[i*verticalGridLineCount+j]=func[i][j];
					}
				}
				java.util.Arrays.sort(value);
				int bottom = 0;
				int top = value.length-1;
				
				while ((max-min)/levelSpacing > maximumLevelCount) {
					if (value[top]-value[top-1] > value[bottom+1]-value[bottom]) {
						top--;
					} else {
						bottom++;
					}
					max = value[top];
					min = value[bottom];
				}
			}
			
			
			// calculate precise levels
			int minLevelIndex = (int)Math.round(
					Math.ceil(min/levelSpacing - levelOffset)
			);
			double minLevel = (levelOffset + minLevelIndex)*levelSpacing;
			int maxLevelIndex = (int)Math.round(
					Math.floor(max/levelSpacing - levelOffset)
			);
			numLevels = maxLevelIndex - minLevelIndex +1;
			for (int j = 0; j<numLevels; j++) {
				double level=minLevel + j*levelSpacing;
				MarchingSquares.IndexedLineSet ls = MarchingSquares.compute(
						func,
						level,
						minX,
						maxX,
						minY,
						maxY
				);
				
				GeneralPath path = new GeneralPath();
				
				int [] length = ls.getLength();
				int [] index = ls.getIndex();
				double [] xy = ls.getPoints();
				
				for( int c=0, k=0; c<length.length; c++ ) {
					
					path.moveTo( (float)xy[2*index[k]], (float)xy[2*index[k]+1] );
					
					k++;
					for( int i=1; i<length[c]; i++, k++ ) {
						path.lineTo( (float)xy[2*index[k]], (float)xy[2*index[k]+1] );
					}
				}
				SceneComponent sg=new SceneComponent();
				sg.setShape(path);
				view.addChild(sg);
			}
			view.fireAppearanceChange();
		}
	}

	public RealFunctionOnRectangle getFunction() {
		return function;
	}

	public void setFunction(RealFunctionOnRectangle function) {
		if (this.function != function) {
			this.function = function;
			updateLevels();
			view.fireAppearanceChange();
		}
	}

	public void functionChanged() {
		updateLevels();
		view.fireAppearanceChange();
	}
	
	public int getHorizontalGridLineCount() {
		return horizontalGridLineCount;
	}

	public void setHorizontalGridLineCount(int horizontalGridLineCount) {
		if (this.horizontalGridLineCount != horizontalGridLineCount) {
			this.horizontalGridLineCount = horizontalGridLineCount;
			func = new double[horizontalGridLineCount][verticalGridLineCount];
			value = new double[horizontalGridLineCount*verticalGridLineCount];
			updateLevels();
			view.fireAppearanceChange();
		}
	}

	public double getLevelOffset() {
		return levelOffset;
	}

	public void setLevelOffset(double levelOffset) {
		if (this.levelOffset != levelOffset) {
			this.levelOffset = levelOffset;
			updateLevels();
			view.fireAppearanceChange();
		}
	}

	public double getLevelSpacing() {
		return levelSpacing;
	}

	public void setLevelSpacing(double levelSpacing) {
		if (this.levelSpacing != levelSpacing) {
			this.levelSpacing = levelSpacing;
			updateLevels();
			view.fireAppearanceChange();
		}
	}

	public int getMaximumLevelCount() {
		return maximumLevelCount;
	}

	public void setMaximumLevelCount(int maximumLevelCount) {
		if (this.maximumLevelCount != maximumLevelCount) {
			this.maximumLevelCount = maximumLevelCount;
			updateLevels();
			view.fireAppearanceChange();
		}
	}

	public Rectangle2D.Double getRectangle() {
		return rectangle;
	}

	public void setRectangle(Rectangle2D.Double rectangle) {
		if (!this.rectangle.equals(rectangle)) {
			this.rectangle = rectangle;
			updateLevels();
			view.fireAppearanceChange();
		}
	}

	public int getVerticalGridLineCount() {
		return verticalGridLineCount;
	}

	public void setVerticalGridLineCount(int verticalGridLineCount) {
		if (this.verticalGridLineCount != verticalGridLineCount) {
			this.verticalGridLineCount = verticalGridLineCount;
			func = new double[horizontalGridLineCount][verticalGridLineCount];
			value = new double[horizontalGridLineCount*verticalGridLineCount];
			updateLevels();
			view.fireAppearanceChange();
		}
	}
	
	public SceneComponent getViewScene() {
		return view;
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}
	
	public Paint getPaint() {
		return view.getOutlinePaint();
	}

	public void setPaint(Paint paint) {
		view.setOutlinePaint(paint);
	}

	public Stroke getStroke() {
		return view.getStroke();
	}
	
	public void setStroke(Stroke stroke) {
		view.setStroke(stroke);
	}
}
