/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.java2dx.numerical;


import java.util.ArrayList;

/**
* For linear subdivision the threshhold length is the maximal edgelength allowed.
* For smooth
* subdivision the threshold length means that an edge e is subdivided if
* one of its endpoints has a distance to a neighboring edge of e that exceeds
* the threshhold length.
* */

public class PolygonSubdivider {
	public final static int LINEAR = 0;
	public final static int CUBIC = 1;
	public final static int CIRCULAR = 2;
	public final static int PARAMETRIZED_CUBIC = 3;
	public final static int BISECTING_CIRCULAR = 4;
	final static double CUTOFF = 1e-10;
	
    static int maxDepth = 10;
    double edgeLengthSquared = Double.MAX_VALUE;
    double curvatureLengthSquared = Double.MAX_VALUE;
    ArrayList points = new ArrayList();
    int subdivisionMethod = CUBIC;
    boolean closed = false;

    public boolean isClosed() {
		return closed;
	}

	public void setClosed(boolean b) {
		closed = b;
	}

	public double[][] subdivide(double[][] p) {
		int m = Integer.MAX_VALUE;
		double[][] q = subdivisionMethod == PARAMETRIZED_CUBIC ?
				constantSpeedParametrization(p) : p;
		int cnt=0;
		while (m!=q.length && cnt<maxDepth) {
			m = q.length;
			if (subdivisionMethod == PARAMETRIZED_CUBIC) {
				if (closed) {
					q = subdivideParametrizedClosedOnce(q);
				} else {
					q = subdivideParametrizedOpenOnce(q);
				}
			} else if (subdivisionMethod == LINEAR){
				q = subdivideLinearOnce(q);
			} else {
				q = subdivideOnce(q);
			}
			cnt++;
		}
		return subdivisionMethod == PARAMETRIZED_CUBIC ?
				forgetParametrization(q) : q;
	}
	
	/**
     * @return Returns the maximal curvature length.
     */
    public double getMaximalEdgeLength() {
        return Math.sqrt(edgeLengthSquared);
    }
    
    /**
     * @param l The maximal curvature length to set.
     */
    public void setMaximalEdgeLength(double l) {
        edgeLengthSquared = l*l;
    }
    
    /**
     * @return Returns the maximal curvature length.
     */
    public double getMaximalCurvatureLength() {
        return Math.sqrt(curvatureLengthSquared);
    }
    
    /**
     * @param l The maximal curvature length to set.
     */
    public void setMaximalCurvatureLength(double l) {
        curvatureLengthSquared = l*l;
    }
    
	public int getSubdivisionMethod() {
		return subdivisionMethod;
	}

	public void setSubdivisionMethod(int subdivisionMethod) {
		this.subdivisionMethod = subdivisionMethod;
	}
    
    double[] twoPointSubdivision(
            double[] z2,
            double[] z3
    ) {
        int n = z2.length;
        double[] result = new double[n];
        for (int j=0; j<n; j++) {
            result[j] = .5 * z2[j] + .5 * z3[j];
        }
        return result;
    }
    
    double[] parametrizedThreePointSubdivision(
            double[] z1,
            double[] z2,
            double[] z3
    ) {
        int n = z1.length;
        double t1 = z1[0];
        double t2 = z2[0];
        double t3 = z3[0];
        double t = (t2+t3)/2;
        double a1 = (t-t2)*(t-t3)/((t1-t2)*(t1-t3));
        double a2 = (t-t1)*(t-t3)/((t2-t1)*(t2-t3));
        double a3 = (t-t1)*(t-t2)/((t3-t1)*(t3-t2));

        double[] result = new double[n];
        result[0] = t;
        for (int j=1; j<n; j++) {
            result[j] = a1*z1[j] + a2*z2[j] + a3*z3[j];
        }
        return result;
    }
    
    double[] fourPointSubdivision(
            double[] z1,
            double[] z2,
            double[] z3,
            double[] z4
    ) {
    	final double c1 = -1/(double)16;
    	final double c2 =  9/(double)16;
        int n = z1.length;
        double[] result = new double[n];
        if (subdivisionMethod == CUBIC) {
        	for (int j=0; j<n; j++) {
        		result[j] = c1*(z1[j]+z4[j]) + c2*(z2[j]+z3[j]);
        	}
        } else {
        	if (subdivisionMethod == CIRCULAR) {
        		double[] mean1 = threePointSubdivision(z1,z2,z3);
        		double[] mean2 = threePointSubdivision(z4,z3,z2);
        		for (int j=0; j<n; j++) {
        			result[j] = .5*(mean1[j] + mean2[j]);
        		}
        	} else {
        		// bisecting circular
        		final double[] v = new double[n];
        		final double[] vInverse = new double[n];
            	final double[] f = new double[n];
            	final double[] g = new double[n];
        		for (int j=0; j<n; j++) {
        			v[j] = z3[j] - z2[j];
        			f[j] = z1[j] - z3[j];
        			g[j] = z4[j] - z3[j];
        			vInverse[j] = v[j];
        		}
        		invert(vInverse);
        		invert(f);
        		invert(g);
        		for (int j=0; j<n; j++) {
        			f[j]+= vInverse[j];
        			g[j]+= vInverse[j];
        		}
        		double normv = Math.sqrt(normSquared(v));
        		double normf = Math.sqrt(normSquared(f));
        		double normg = Math.sqrt(normSquared(g));
        		for (int j=0; j<n; j++) {
        			f[j] = -(f[j]/normf + g[j]/normg);
        		}
        		normf = Math.sqrt(normSquared(f))*normv;
        		for (int j=0; j<n; j++) {
        			f[j] = f[j]/normf - vInverse[j];
        		}
        		invert(f);
        		for (int j=0; j<n; j++) {
        			result[j] = f[j] + v[j]+z2[j];
        		}
        	}
        }
        return result;
    }
    
    double[] parametrizedFourPointSubdivision(
            double[] z1,
            double[] z2,
            double[] z3,
            double[] z4
    ) {
    	int n = z1.length;
    	double[] result = new double[n];
    	double t1 = z1[0];
    	double t2 = z2[0];
    	double t3 = z3[0];
    	double t4 = z4[0];
    	double t = (t2+t3)/2;
    	double a1 = (t-t2)*(t-t3)*(t-t4)/((t1-t2)*(t1-t3)*(t1-t4));
    	double a2 = (t-t1)*(t-t3)*(t-t4)/((t2-t1)*(t2-t3)*(t2-t4));
    	double a3 = (t-t1)*(t-t2)*(t-t4)/((t3-t1)*(t3-t2)*(t3-t4));
    	double a4 = (t-t1)*(t-t2)*(t-t3)/((t4-t1)*(t4-t2)*(t4-t3));
    	result[0] = t;
    	for (int j=1; j<n; j++) {
    		result[j] = a1*z1[j] + a2*z2[j] + a3*z3[j] + a4*z4[j];
    	}
    	return result;
    }
    
    double[] threePointSubdivision(double[] a0, double[] a1,double[] a2) {
    	final double c0 = -1/(double)8;
    	final double c1 = 3/(double)4;
    	final double c2 = 3/(double)8;
    	int n = a0.length;
    	final double[] f = new double[n];
    	if (subdivisionMethod == CUBIC ||
    			subdivisionMethod == PARAMETRIZED_CUBIC) {
    		for (int j=0; j<n; j++) {
    			f[j] = c0*a0[j] + c1*a1[j] + c2*a2[j];
    		}
    	} else {
    		final double[] v = new double[n];
    		final double[] vInverse = new double[n];
    		
    		for (int j=0; j<n; j++) {
    			v[j] = a2[j] - a1[j];
    			f[j] = a0[j] - a2[j];
    			vInverse[j] = v[j];
    		}
    		invert(vInverse);
    		invert(f);
    		for (int j=0; j<n; j++) {
    			f[j]+= vInverse[j];
    		}
    		double norm = Math.sqrt(normSquared(f))*Math.sqrt(normSquared(v));
    		for (int j=0; j<n; j++) {
    			f[j] = -f[j]/norm - vInverse[j];
    		}
    		invert(f);
    		for (int j=0; j<n; j++) {
    			f[j] += v[j]+a1[j];
    		}
    	}
    	return f;
    }
    
    void invert(double[] v) {
    	double absSquared = normSquared(v);
    	int n = v.length;
    	for (int j=0; j<n; j++) {
    		v[j]/=absSquared;
    	}
    }
    
    double normSquared(double[] v) {
    	int n = v.length;
    	double norm = 0;
    	for (int j=0; j<n; j++) {
    		norm += v[j]*v[j];
    	}
    	return norm;
    }
    double[][] forgetParametrization(double[][] p) {
    	int n = p.length;
    	int m = p[0].length;
    	double[][] result = new double[n][m-1];
    	for (int j=0; j<n; j++) {
    		System.arraycopy(p[j],1,result[j],0,m-1);
    	}
    	return result;
    }
    
    double[][] constantSpeedParametrization(
            double[][] p
    ) {
        int n=p.length;
        int m = p[0].length;
        double[][] result = new double[n][m+1];
        double length = 0;
        for (int j=0; j<n-1; j++) {
            result[j][0] = length;
            System.arraycopy(p[j],0,result[j],1,m);
            double edgeLength = 0;
            for (int i=0; i<m; i++) {
                double a = p[(j+1)][i]-p[j][i];
                edgeLength += a*a;
            }
            length += Math.sqrt(edgeLength);
        }
        System.arraycopy(p[n-1],0,result[n-1],1,m);
        result[n-1][0]=length;
        if (closed) {
            // add the last edge
            double edgeLength = 0;
            for (int i=0; i<m; i++) {
                double a = p[n-1][i]-p[0][i];
                edgeLength += a*a;
            }
            length += Math.sqrt(edgeLength);
        }
        
        // normalize parameter values for total length 1
        double factor = 1/length;
        for (int j=0; j<n; j++) {
            result[j][0]*=factor;
        }
        return result;
    }
    
    double[][] coarsenClosedOnce(double[][] p) {
        int n=p.length;
        int m = p[0].length;
        points.clear();
        double[] edge1 = new double[m-1];
        double[] edge2 = new double[m-1];
        double[] edge3 = new double[m-1];
        
        double[] q1 = new double[m];
        double[] q2 = new double[m];
        double[] r1,r2,r3,r4;
        for (int j=0; j<n; j++) {
            
            if (j<2) {
                System.arraycopy(p[n+j-2],0,q1,0,m);
                q1[0] -= 1;
                r1 = q1;
            } else {
                r1 = p[j-2];
            }
            
            if (j<1) {
                System.arraycopy(p[n+j-1],0,q1,0,m);
                q1[0] -= 1;
                r2 = q1;
            } else {
                r2 = p[j-1];
            }
            
            if (j>n-2) {
              System.arraycopy(p[j-n+1],0,q1,0,m);
              q1[0]+=1;
              r3=q1;
            } else {
                r3 = p[j+1];
            }
              
            if (j>n-3) {
                System.arraycopy(p[j-n+2],0,q2,0,m);
                q2[0]+=1;
                r4 = q2;
            } else {
                r4 = p[j+2];
            }
            
            for (int i = 1; i<m; i++) {
                edge1[i-1] = r2[i]-r1[i];
                edge2[i-1] = r3[i]-r2[i];
                edge3[i-1] = r4[i]-r3[i];
            }
            if (!isTooLong(edge2,edge1) && !isTooLong(edge2,edge3)) {
                double[]q = new double[m];
                System.arraycopy(p[j],0,q,0,m);
                points.add(q);
            } else {
                //found an unnecessary vertex
                for (int k=j+1; k<n; j++) {
                    double[]q = new double[m];
                    System.arraycopy(p[k],0,q,0,m);
                    points.add(q);
                }
                break;
            }
        }
        int r = points.size();
        double[][] result = new double[r][];
        for (int i=0; i<r; i++) {
            result[i]=(double[])points.get(i);
        }
        return result;
    }
    
    double[][] subdivideLinearOnce(double[][] p) {
    	int n=p.length;
        int m = p[0].length;
        points.clear();
        double[] r2,r3;
        for (int j=0; j<n; j++) {
            double[]q = new double[m];
            System.arraycopy(p[j],0,q,0,m);
            points.add(q);
            if (!closed && j==n-1) break;  
            r2 = p[j];
            r3 = p[(j+1)%n];
            double edgeLength = 0;
            for (int i=0; i<m; i++) {
                double d = r3[i]-r2[i];
                edgeLength += d*d;
            }
            if (edgeLength > edgeLengthSquared) {
                points.add(twoPointSubdivision(r2,r3));
            }
        }
        // collect all the points
        int r = points.size();
        double[][] result = new double[r][];
        for (int i=0; i<r; i++) {
            result[i]=(double[])points.get(i);
        }
        return result;
    }
    
    double[][] subdivideParametrizedClosedOnce(double[][] p) {
        int n=p.length;
        int m = p[0].length;
        points.clear();
        double[] edge1 = new double[m-1];
        double[] edge2 = new double[m-1];
        double[] edge3 = new double[m-1];
        
        double[] q1 = new double[m];
        double[] q2 = new double[m];
        double[] r1,r2,r3,r4;
        for (int j=0; j<n; j++) {
            double[]q = new double[m];
            System.arraycopy(p[j],0,q,0,m);
            points.add(q);
            
            if (j==0) {
                System.arraycopy(p[n-1],0,q1,0,m);
                q1[0] -= 1;
                r1 = q1;
            } else {
                r1 = p[j-1];
            }
            
            r2 = p[j];

            if (j==n-1) {
              System.arraycopy(p[0],0,q1,0,m);
              q1[0]+=1;
              r3=q1;
            } else {
                r3 = p[j+1];
            }
              
            
            int i4 = (j+2)%n;
            if (j>n-3) {
                System.arraycopy(p[i4],0,q2,0,m);
                q2[0]+=1;
                r4 = q2;
            } else {
                r4 = p[j+2];
            }
            
            for (int i = 1; i<m; i++) {
                edge1[i-1] = r2[i]-r1[i];
                edge2[i-1] = r3[i]-r2[i];
                edge3[i-1] = r4[i]-r3[i];
            }
            if (isTooLong(edge2,edge1) || isTooLong(edge2,edge3)) {

                points.add(parametrizedFourPointSubdivision(r1,r2,r3,r4));
            }
        }
        int r = points.size();
        double[][] result = new double[r][];
        for (int i=0; i<r; i++) {
            result[i]=(double[])points.get(i);
        }
        return result;
    }
    
    double[][] subdivideOnce(double[][] p) {
        int n=p.length;
        int m = p[0].length;
        points.clear();
        double[] edge1 = new double[m];
        double[] edge2 = new double[m];
        double[] edge3 = new double[m];
        double[] r1,r2,r3,r4;
        int fourPointStart = closed ? 0 : 1;
        int fourPointEnd = closed ? n : n-2;
        if (!closed) {
        	// deal with first edge
        	double[]q = new double[m];
            System.arraycopy(p[0],0,q,0,m);
            points.add(q);
        	for (int i = 0; i<m; i++) {
                edge1[i] = p[2][i] - p[1][i];
            	edge2[i] = p[1][i] - p[0][i];
            }
        	if (isTooLong(edge2,edge1)) {
        		points.add(threePointSubdivision(p[2],p[1],p[0]));
        	}
        }
        for (int j=fourPointStart; j<fourPointEnd; j++) {
        	// add the old point
            double[]q = new double[m];
            System.arraycopy(p[j],0,q,0,m);
            points.add(q);
            r1 = p[j == 0 ? n-1 : j-1];      
            r2 = p[j];
            r3 = p[(j+1)%n];
            r4 = p[(j+2)%n]; 
            for (int i = 0; i<m; i++) {
            	edge1[i] = r2[i]-r1[i];
                edge2[i] = r3[i]-r2[i];
                edge3[i] = r4[i]-r3[i];
            }
            if (isTooLong(edge2,edge1) || isTooLong(edge2,edge3)) {
                points.add(fourPointSubdivision(r1,r2,r3,r4));
            }
        }
        if (!closed) {
        	// deal with last edge
        	double[]q = new double[m];
            System.arraycopy(p[n-2],0,q,0,m);
            points.add(q);
        	for (int i = 0; i<m; i++) {
                edge1[i] = p[n-3][i] - p[n-2][i];
            	edge2[i] = p[n-2][i] - p[n-1][i];
            }
        	if (isTooLong(edge2,edge1)) {
        		points.add(threePointSubdivision(p[n-3],p[n-2],p[n-1]));
        	}
        	q = new double[m];
            System.arraycopy(p[n-1],0,q,0,m);
            points.add(q);
        }
        // collect all the points
        int r = points.size();
        double[][] result = new double[r][];
        for (int i=0; i<r; i++) {
            result[i]=(double[])points.get(i);
        }
        return result;
    }
    
    double[][] subdivideParametrizedOpenOnce(double[][] p) {
        int n=p.length;
        int m = p[0].length;
        points.clear();
        double[] edge1 = new double[m-1];
        double[] edge2 = new double[m-1];
        double[] edge3 = new double[m-1];
        double[] r1,r2,r3,r4;
        
        double[]q = new double[m];
        System.arraycopy(p[0],0,q,0,m);
        points.add(q);
        
        r2 = p[0];
        r3 = p[1];
        r4 = p[2];
        for (int i = 1; i<m; i++) {
            edge2[i-1] = r3[i]-r2[i];
            edge3[i-1] = r4[i]-r3[i];
        }
        if (isTooLong(edge2,edge3)) {
            points.add(threePointSubdivision(r2,r3,r4));
        }
        
        for (int j=1; j<n-1; j++) {
            q = new double[m];
            System.arraycopy(p[j],0,q,0,m);
            points.add(q);
            
            r1 = p[j-1];
            r2 = p[j];
            r3 = p[j+1];
            r4 = p[j+2];
            
            for (int i = 1; i<m; i++) {
                edge1[i-1] = r2[i]-r1[i];
                edge2[i-1] = r3[i]-r2[i];
                edge3[i-1] = r4[i]-r3[i];
            }
            
            if (isTooLong(edge2,edge1) || isTooLong(edge2,edge3)) {
                points.add(fourPointSubdivision(r1,r2,r3,r4));
            }
        }
        // TODO: n-1
        int r = points.size();
        double[][] result = new double[r][];
        for (int i=0; i<r; i++) {
            result[i]=(double[])points.get(i);
        }
        return result;
    }
    
    boolean isTooLong(double[] e, double[] f) {
        int n = e.length;
        double ne=0;
        double nf=0;
        double s=0;
        for (int i=0; i<n; i++) {
            ne += e[i]*e[i];
            nf += f[i]*f[i];
            s += e[i]*f[i];
        }
        return ne - s*s/nf > curvatureLengthSquared || ne > edgeLengthSquared;
    }

    double[][] coarsenClosed(double[][] p) {
        int m = Integer.MAX_VALUE;
        double[][] q = p;
        int cnt=0;
        while (m!=q.length && q.length>3) {
            m = q.length;
            q = coarsenClosedOnce(q);
            cnt++;
        }
        return q;
    }
}
