/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.java2dx.plugin;

import java.awt.geom.Rectangle2D;
import java.util.Collections;
import java.util.Set;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import de.jtem.beans.InspectorPanel;
import de.jtem.java2d.Viewer2D;
import de.jtem.java2d.ViewportChangeEvent;
import de.jtem.java2d.ViewportChangeListener;
import de.jtem.java2dx.numerical.ContourPlot;
import de.jtem.java2dx.numerical.RealFunctionOnRectangle;
import de.jtem.java2dx.plugin.icon.ImageHook;
import de.jtem.jrworkspace.plugin.Controller;
import de.jtem.jrworkspace.plugin.PluginInfo;
import de.jtem.jrworkspace.plugin.sidecontainer.SideContainerPerspective;
import de.jtem.jrworkspace.plugin.sidecontainer.template.ShrinkPanelPlugin;
import de.jtem.jrworkspace.plugin.simplecontroller.SimpleController;

public class ContourPlotPlugin extends ShrinkPanelPlugin {

	final static Set<String> excludedPropertiesForContourPlot = Collections.singleton("name");
	private ContourPlot plot = new ContourPlot();
	private Viewer2D viewer;
	private boolean contourPlotEnabled = false;
	private ViewportChangeListener plotListener = new ViewportChangeListener() {

		public void viewportChange(ViewportChangeEvent e) {
			plot.setRectangle(viewer.getViewport());
		}
	};
	private InspectorPanel inspectorPanel;

	public ContourPlotPlugin() {
		super();
		inspectorPanel = new InspectorPanel(false);
		getShrinkPanel().add(inspectorPanel);
	}

	@Override
	public Class<? extends SideContainerPerspective> getPerspectivePluginClass() {
		return Java2DView.class;
	}

	public PluginInfo getPluginInfo() {
		PluginInfo info = new PluginInfo();
		info.name = "Contour Plot";
		info.vendorName = "Ulrich Pinkall"; 
		info.icon = ImageHook.getIcon("wrench.png");
		info.isDynamic = false;
		return info;
	}

	@Override
	public void install(Controller c) throws Exception {
		Java2DView view = c.getPlugin(Java2DView.class);
		viewer = view.getViewer2D();
		viewer.addViewportChangeListener(new ViewportChangeListener() {

			public void viewportChange(ViewportChangeEvent e) {
				plot.setRectangle(viewer.getViewport());
			}
		});
		inspectorPanel.setObject(plot);
		inspectorPanel.addChangeListener(
				new ChangeListener() {
    				public void stateChanged(ChangeEvent event) {
    					if (event.getSource() == plot) {
    						viewer.repaint();
    					}
    				}
    			}
    	);
		setContourPlotEnabled(true);
		super.install(c);
	}

	@Override
	public void uninstall(Controller c) throws Exception {
		inspectorPanel.setObject(null);
		super.uninstall(c);
	}
	
	public void setContourPlotEnabled(boolean b) {
		if (b && !contourPlotEnabled) {
			viewer.addViewportChangeListener(plotListener);
			viewer.getBackdrop().addChild(plot.getViewScene());
			contourPlotEnabled = true;
			plot.setRectangle(viewer.getViewport());
			plot.functionChanged();
			
			viewer.repaint();
		}
		if (!b && contourPlotEnabled) {
			viewer.removeViewportChangeListener(plotListener);
			viewer.getBackdrop().removeChild(plot.getViewScene());
			contourPlotEnabled = false;
			plot.functionChanged();
			viewer.repaint();
		}
	}
	
	public static void main(String[] args) {
		
		SimpleController c = new SimpleController();
		final Java2DView view = new Java2DView();
		view.getViewer2D().setInitialViewPort(new Rectangle2D.Double(-5,-5,10,10));
		c.registerPlugin(view);
		c.registerPlugin(new Java2DViewMenuBar());
		c.registerPlugin(new Viewer2DInspector());
		c.registerPlugin(new GridInspector());
		
		ContourPlotPlugin plot = new ContourPlotPlugin();
		RealFunctionOnRectangle f = new RealFunctionOnRectangle() {
			
			public double value(double x, double y) {
				return Math.cos(x)+Math.sin(y);
			}
			
		};
		plot.getContourPlot().setFunction(f);
		c.registerPlugin(plot);
		
		c.startup();
	}

	public ContourPlot getContourPlot() {
		return plot;
	}
}
